from typing import List, Callable
from .get_annotation_fn import crossval


class GetAnnotationABC:
    target: str

    # for train annotation
    img_dir_train: str
    label_dir_train: str
    train_dirs: List[str] = None
    get_train_fn: Callable[[str, str, str, List[str]], List[List[str]]]

    # for val annotation
    img_dir_val: str = None
    label_dir_val: str = None
    val_dirs: List[str] = None
    get_val_fn: Callable[[str, str, str, List[str]], List[List[str]]] = None

    @classmethod
    def __call__(cls, cv_fold: int = None, cv_i=0, depth=0):
        if cv_fold:
            print(f"cross_val: {cv_i+1}/{cv_fold}")
            return crossval(cls.get_train_annotations(), cv_fold, cv_i, depth)
        else:
            return cls.get_train_annotations(), cls.get_val_annotations()

    @classmethod
    def get_train_annotations(cls):
        return cls.get_train_fn(
            cls.target, cls.img_dir_train, cls.label_dir_train, cls.train_dirs)

    @classmethod
    def get_val_annotations(cls):
        return cls.get_val_fn(
            cls.target, cls.img_dir_val, cls.label_dir_val, cls.val_dirs)



from typing import List
import collections
import numpy as np
import re
from pathlib import Path


def seg_case_direct(
      root: str,
      image_dir: str,
      label_dir: str,
      *args
      ) -> List[List[Path]]:

    """
    - root
        - image_dir
        - label_dir
    """

    annotations = list()
    c_label, c_img = Path(root) / label_dir, Path(root) / image_dir
    labels = sorted(_get_img_files(c_label))
    imgs = [next(c_img.glob(f"{label.stem}.*")) for label in labels]
    annotations = list(zip(imgs, labels))
    return annotations


def seg_case_first_targets(
      root: str,
      image_dir: str,
      label_dir: str,
      target_dirs: List[str]
      ) -> List[List[Path]]:

    """
    caseごとにフォルダが作成されている場合
    - root
        - case_name
            - image_dir
            - label_dir
    target_dirs: [case_name1, case_name2, ...]
    """
    annos = list()
    for case_name in target_dirs:
        case_dir = Path(root) / case_name
        annos += seg_case_direct(str(case_dir), image_dir, label_dir)
    return annos


def seg_case_first_groups(
      root: str,
      image_dir: str,
      label_dir: str,
      group_dirs: List[str]
      ) -> List[List[Path]]:

    """
    caseごとのフォルダをさらにグループでまとめている場合
    - root
        - group_name
            - case_name
                - image_dir
                - label_dir
    group_dirs: [group_name1, group_name2, ...]
    """

    annos = list()
    for group_name in group_dirs:
        group_dir = Path(root) / group_name
        for case_dir in group_dir.iterdir():
            annos += seg_case_direct(str(case_dir), image_dir, label_dir)
    return annos


def _get_img_files(p_dir: Path) -> List[Path]:
    ImageEx = "jpg|jpeg|png|gif|bmp"
    img_files = [
        p for p in p_dir.glob('*') if re.search(rf'.*\.({ImageEx})', str(p))
    ]
    return img_files


def crossval(annos: List[List[Path]], cv_fold, cv_i, depth):
    if depth == 0:
        nb_val = len(annos) // cv_fold
        train_annos = annos[:nb_val*cv_i] + annos[nb_val*(cv_i+1):]
        val_annos = annos[nb_val*cv_i:nb_val*(cv_i+1)]
    else:
        group_names = list()
        for img_file, _ in annos:
            for _ in range(depth+1):
                img_file = img_file.parent
            group_names.append(img_file.stem)
        group_counter = collections.Counter(group_names)
        cross_val_dirs = _cross_val_split(group_counter, cv_fold)
        val_dirs = cross_val_dirs[cv_i]
        train_dirs = list()
        for val_i in range(cv_fold):
            if val_i == cv_i:
                continue
            train_dirs += cross_val_dirs[val_i]
        train_annos = [
            anno for anno, group_name in zip(annos, group_names)
            if group_name in train_dirs
        ]
        val_annos = [
            anno for anno, group_name in zip(annos, group_names)
            if group_name in val_dirs
        ]
    return train_annos, val_annos


def _cross_val_split(g_count, k=5, n_iter=15, mix_step=5):
    data = [dict(case=case, count=count) for case, count in g_count.items()]
    # sort raughly
    data = sorted(data, reverse=True, key=lambda x: x["count"])
    cross = [[v] for v in data[:k]]
    for v in data[k:]:
        sum_list = [sum([case_dic["count"] for case_dic in c]) for c in cross]
        min_sum_id = np.argmin(sum_list)
        cross[min_sum_id].append(v)

    # n swap sort
    swapping = False
    for step in range(n_iter):
        for i in range(k-1):
            for j in range(k-i):
                n = cross[-j-1]
                m = cross[i]
                n_sum = sum([case_dic["count"] for case_dic in n])
                m_sum = sum([case_dic["count"] for case_dic in m])
                if n_sum > m_sum:
                    large = n
                    small = m
                    diff = n_sum - m_sum
                else:
                    small = n
                    large = m
                    diff = m_sum - n_sum

                for i_l, v_l in enumerate(large):
                    for i_s, v_s in enumerate(small):
                        no_change_swap = int((step + 1) % mix_step == 0)
                        l_s_diff = v_l["count"] - v_s["count"]
                        if 0 < l_s_diff < diff + no_change_swap:
                            large[i_l] = v_s
                            small[i_s] = v_l
                            swapping = True
                            break
                    if swapping:
                        swapping = False
                        break

    cross_val_dirs = list()
    for case_list in cross:
        val_dirs = [d["case"] for d in case_list]
        cross_val_dirs.append(val_dirs)

    return cross_val_dirs



import torch
from pathlib import Path
from .get_optimization_fn import get_prog_bar


class GetOptimizationABC:
    batch_size: int
    epochs: int
    lr: float
    gpu: int
    optim_obj: torch.optim.Optimizer

    def __init__(self, model, loss_func, metrics, train_dataset, val_dataset):
        self.model = model
        self.loss_func = loss_func
        self.metrics = metrics
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset

    def __call__(self):

        train_loader = torch.utils.data.DataLoader(
            self.train_dataset, batch_size=self.batch_size, shuffle=True)
        valid_loader = torch.utils.data.DataLoader(
            self.val_dataset, batch_size=self.batch_size, shuffle=False)

        device = torch.device(
            f"cuda:{self.gpu}" if torch.cuda.is_available() else "cpu")
        print(device)
        self.model.to(device)

        self.optimizer = self.optim_obj(
            [dict(params=self.model.parameters(), lr=self.lr)])

        save_model_dir = Path("./models")
        save_model_dir.mkdir(exist_ok=True)

        for epoch in range(self.epochs):
            print(f"\ntrain step, epoch: {epoch + 1}/{self.epochs}")
            self.model.train()
            training_loss, training_dice = 0, 0
            nb_iters = len(train_loader.dataset) // self.batch_size + 1
            for i, data in enumerate(train_loader, 1):
                # get the inputs; data is a list of [inputs, labels]
                inputs, labels = data
                inputs, labels = inputs.to(device), labels.to(device)

                outputs = self.model(inputs)
                loss = self.loss_func(outputs, labels)

                # zero the parameter gradients
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                # print statistics
                training_loss += loss.item()
                training_dice += self.metrics(outputs, labels).item()
                cout = get_prog_bar(i, nb_iters)
                cout += f" loss: {(training_loss / i):.5g}"
                cout += f" dice: {(training_dice / i):.5g}"
                print("\r"+cout, end="")

            # validation step
            print("\nvalidation step")
            self.model.eval()
            total_loss, total_dice = 0, 0
            with torch.no_grad():
                nb_iters = len(valid_loader.dataset) // self.batch_size + 1
                for i, data in enumerate(valid_loader, 1):
                    inputs, labels = data
                    inputs, labels = inputs.to(device), labels.to(device)
                    outputs = self.model(inputs)
                    total_loss += self.loss_func(outputs, labels).item()
                    total_dice += self.metrics(outputs, labels).item()

                    cout = get_prog_bar(i, nb_iters)
                    val_loss, val_dice = total_loss / i, total_dice / i
                    cout += f" val_loss: {val_loss:.5g}"
                    cout += f" val_dice: {val_dice:.5g}"
                    print("\r"+cout, end="")

            model_path = f'{save_model_dir}/model_epoch{epoch}.pth'
            torch.save(self.model.state_dict(), model_path)

            self.on_epoch_end()  # custom callbacks

        print('\nFinished Training')
        return val_dice

    def on_epoch_end(self):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['farmer_pytorch',
 'farmer_pytorch.GetAnnotation',
 'farmer_pytorch.GetDataset',
 'farmer_pytorch.GetOptimization']

package_data = \
{'': ['*'],
 'farmer_pytorch': ['Util/Cls/postprocess/*',
                    'Util/Cls/preprocess/*',
                    'Util/Det/postprocess/*',
                    'Util/Det/preprocess/*',
                    'Util/Segm/postprocess/*',
                    'Util/Segm/preprocess/*']}

install_requires = \
['Pillow>=8.3.2,<9.0.0', 'numpy>=1.21.2,<2.0.0', 'torch>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'farmer-pytorch',
    'version': '0.1.3',
    'description': 'deep learning tools: easy to run, easy to customize',
    'long_description': '# Pytorch segmentation\n\n## installation\n```\npip install farmer-pytorch\n```\n\n\n## Quick start\n[quick start](quickstart/)',
    'author': 'aiorhiroki',
    'author_email': '1234defgsigeru@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aiorhiroki/farmer.pytorch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



# プログレスバーの作成
def get_prog_bar(iter_i, iter_total, length=50):
    prob_norm = length / iter_total
    done = '=' * int(iter_i * prob_norm)
    todo = ' ' * int((iter_total - iter_i) * prob_norm)
    prob_bar = done+todo
    return f"[{prob_bar}] {iter_i}/{iter_total}"



from torch.utils.data import Dataset
from PIL import Image
import numpy as np
from typing import List


class GetDatasetSgmABC(Dataset):
    class_values: List[int]

    def __init__(self, annotation, augmentation):
        self.annotation = annotation
        self.augmentation = augmentation

    def __getitem__(self, i):
        img_file, label_file = self.annotation[i]

        image = np.array(Image.open(str(img_file)))
        mask = np.array(Image.open(str(label_file)))

        # apply augmentations
        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample['image'], sample['mask']

        # custom preprocessing
        image, mask = self.preprocess(image, mask)

        # preprocess image for input
        image = image.transpose(2, 0, 1).astype('float32') / 255.

        masks = [(mask == v) for v in self.class_values]
        mask = np.array(masks, dtype='float32')

        return image, mask

    def __len__(self):
        return len(self.annotation)

    def preprocess(self, image, mask):
        return image, mask
